#! /usr/bin/python

import copy
import string
import plistlib
import itertools
import sys

from PIL import Image

import array
from math import *
from sprite import *
from spritetile import *
from tile import *
from tilificator import *
from TilificatorProject import TilificatorProject
from import_raw_tiles import *
from tilingmethodshiftedrows import TilingMethodShiftedRows
from tilingmethoddragqueen import TilingMethodDragQueen
from tilingmethodfullmatch import TilingMethodFullMatch

def blankLine(st, line, si):
    for xx in range(st.w):
        x = st.x+xx
        y = st.y+line
        if 0 <= x < si.width and 0 <= y < si.height and (si.data[y*si.width+x] & 0x3):
            return False
    return True

def findFirstLineST(st, si):
    for i in range(8):
        if not blankLine(st, i, si):
            return st.y+i

def findFirstLine(si):
    firstLine = 7
    for st in si.tilification.tiles:
        firstLine = min(firstLine,findFirstLineST(st, si))
    return firstLine

def bankOfSpriteImage(spriteImageIndex, bankStart):
    for bank,startIndex in reversed(list(enumerate(bankStart))):
        if startIndex <= spriteImageIndex:
            return bank
    return len(bankStart)

def removeDuplicates(tiles):
    usedTiles = []
    for st in tiles:
        if st.tileID not in usedTiles:
            usedTiles.append(st.tileID)
    return usedTiles
#
# Run as stand-alone program
#
if __name__ == "__main__":
    if len(sys.argv) < 2:
        print "Usage: convertSpritesNES.py <tilificatorproject.tpr>" # <sprites.asm> <tiledata.chr>"
        sys.exit(2)
    spriteImages = []
    bankStart = [0]
    for bank,filename in enumerate(sys.argv[1:]):
        with open(filename) as projectFile:
            tpr = TilificatorProject()
            tpr.read(projectFile)
            spriteImages += tpr.spriteImages
        bankStart.append(len(spriteImages))
    # Write .asm file
    f = sys.stdout
    for i,si in enumerate(spriteImages):
        f.write('SpriteImageAddr%d:\n' % i)
        f.write('.byte %d\n' % len(si.tilification.tiles))
        usedTiles = removeDuplicates(si.tilification.tiles)
        for st in si.tilification.tiles:
            H = 1 if st.flipH else 0
            V = 1 if st.flipV else 0
            tileID = ((11-usedTiles.index(st.tileID)) & 0xFF) if i >= bankStart[1] else st.tileID
            f.write('.byte %d,%d,%d,%d\n' % (st.y+st.h-findFirstLine(si),tileID,(V<<7)|(H<<6),st.x+st.w))
            
    f.write('SpriteImageAddrTabLo:\n')
    for i,si in enumerate(spriteImages):
        f.write('.byte <SpriteImageAddr%d\n' % i)
    f.write('\n')
    f.write('SpriteImageAddrTabHi:\n')
    for i,si in enumerate(spriteImages):
        f.write('.byte >SpriteImageAddr%d\n' % i)
    f.write('\n')

    # Write TilesAddr to file
    with open('dragon_tilesref.inc', 'w') as f:
        for i,si in enumerate(spriteImages):
            # if bank 1, write tiles addr
            if i >= bankStart[1]:
                f.write('TilesRef%d:\n.byte ' % i)
                usedTiles = removeDuplicates(si.tilification.tiles)
                for j,tileID in enumerate(usedTiles):
                    f.write(str(tileID))
                    if j == len(usedTiles)-1:
                        f.write('\n')
                    else:
                        f.write(',')

        f.write('.define TilesRefAddr ')        
        for i,si in enumerate(spriteImages):
            if i != 0:
                f.write(',')
            if i >= bankStart[1]:
                f.write('TilesRef%d' % i)
            else:
                f.write('0')
        f.write('\nTileRefAddrLo: .lobytes TilesRefAddr\n')
        f.write('TileRefAddrHi: .hibytes TilesRefAddr\n')

        f.write('TilesPage:\n.byte ')
        for i,si in enumerate(spriteImages):
            if i != 0:
                f.write(',')
            f.write('%d' % bankOfSpriteImage(i,bankStart))

