.SETCPU "6502X"
.IMPORTZP frameCounter,updateState

.include "actor.h"
.include "common.h"
.include "input.h"
.include "gfx.h"
.include "map.h"
.include "soundcode.h"
.include "serial.h"

SPEED_SCALE = 5    ; 5 for NTSC, 6 for PAL

MAX_ACTORS = 16

ACTOR_DIRECTION     = %01000000
ACTOR_PRIORITY      = %00011000
ACTOR_PALETTE       = %00000011

.MACRO ACTOR_SET_PALETTE palette
    lda actorFlagsZP
    and #<~ACTOR_PALETTE
    ora palette
    sta actorFlagsZP
.ENDMACRO

.MACRO ACTOR_SET_PRIORITY priority
    lda actorFlagsZP
    and #<~ACTOR_PRIORITY
    ora #(8*priority)
    sta actorFlagsZP
.ENDMACRO

.MACRO ACTOR_SET_PRIORITY_SIBLING priority
    lda actorFlags,x
    and #<~ACTOR_PRIORITY
    ora #(8*priority)
    sta actorFlags,x
.ENDMACRO

.ZEROPAGE
actorFlagsZP:       .res    1
actorTileOffset:    .res    2
preloadAddrMask:    .res    2
preloadSpriteImage: .res    2

.SEGMENT "ActorRAM"
actorType:          .res    MAX_ACTORS
actorFlags:         .res    MAX_ACTORS
actorLoX:           .res    MAX_ACTORS
actorHiX:           .res    MAX_ACTORS
actorLoY:           .res    MAX_ACTORS
actorHiY:           .res    MAX_ACTORS
actorAddrLo:        .res    MAX_ACTORS
actorAddrHi:        .res    MAX_ACTORS
actorSpriteImage:   .res    MAX_ACTORS
actorAnimation:     .res    MAX_ACTORS
actorAnimationPos:  .res    MAX_ACTORS
actorFrameDelay:    .res    MAX_ACTORS

; Actor-specific data
actorData0:         .res    MAX_ACTORS
actorData1:         .res    MAX_ACTORS
actorData2:         .res    MAX_ACTORS
actorData3:         .res    MAX_ACTORS

; Actor classes, controlled by setting ACTORCLASS = xyz_ACTOR
.define DEFAULT_ACTOR               0
.define BLOCK_ACTOR                 1
.define BLOCK_ACTOR_ATTACHED_TOP    2
.define CONSTRUCTION_ACTOR          3

actorMapPos = actorData0

.CODE
.include "animations.inc"

.SEGMENT "ACTORCODE"
.MACRO ACTOR_FACING_RIGHT
    bit actorFlagsZP
    set_flag_test V clear
.ENDMACRO

.MACRO ACTOR_FACING_LEFT
    bit actorFlagsZP
    set_flag_test V set
.ENDMACRO

.MACRO INCDIR_Y
    jsr IncDirYSub
.ENDMACRO

.MACRO DECDIR_Y
    jsr DecDirYSub
.ENDMACRO

YieldSub:
    pla
    sta actorAddrLo,x
    pla
    sta actorAddrHi,x
    lda actorFlagsZP
    sta actorFlags,x
    rts

.MACRO YIELD
.IF (ACTORCLASS = BLOCK_ACTOR)
    jsr BlockActorYieldSub
.ELSEIF (ACTORCLASS = BLOCK_ACTOR_ATTACHED_TOP)
    jsr BlockActorYieldSub
    jsr ReadjustBlockActorAttachedTop
.ELSEIF (ACTORCLASS = CONSTRUCTION_ACTOR)
    jsr ConstructionActorYieldSub
    jsr ReadjustY
.ELSE
    jsr AnimatedActorYieldSub
.ENDIF
.ENDMACRO

KillSelfSub:
    lda #0
    sta actorAddrHi,x
    lda #ACTOR_TYPE_NONE
    sta actorType,x
    rts

.MACRO KILL_SELF
    jmp KillSelfSub
.ENDMACRO

SetAnimationSub:
    sty TEMP+2
    sta actorAnimation,x
    lda #4
    sta actorAnimationPos,x
    
    ldy actorAnimation,x
    lda SprAnimAddrLo,y
    sta TEMP
    lda SprAnimAddrHi,y
    sta TEMP+1
    ldy actorAnimationPos,x
    iny
    lda (TEMP),y
    sec
    sbc #1
    sta actorFrameDelay,x
    ; Preload next image if this is actor #0 or #1
    cpx #2
    bpl :+
    jsr PreloadSpriteImage
:
    ldy TEMP+2
    rts

;
; Adjust carry depending on frame counter and player
; to make animation frames advance on even frames
; for player0 and odd frames for player1
;
AdjustCarry:
    txa
    eor #1
    eor frameCounter
    lsr
    rts
    
.MACRO ACTOR_SET_ANIMATION animation
    lda animation
    jsr SetAnimationSub
.ENDMACRO

CheckAnimationDoneSub:
    ldy actorAnimation,x
    lda actorAnimationPos,x
    cmp SprAnimEnd,y
    rts

.MACRO CHECK_ANIMATION_DONE
    jsr CheckAnimationDoneSub
.ENDMACRO

.macro WAIT_UNTIL_ANIMATION_DONE
    .LOCAL WaitLoop
    WaitLoop:
        YIELD
        CHECK_ANIMATION_DONE
    bne WaitLoop ;while not equal
.endmacro

AdvanceAnimationSub:
    dec actorFrameDelay,x
    beq @AdvanceFrame
    rts
@AdvanceFrame:
    lda actorAnimationPos,x
    clc
    adc #4
    sta actorAnimationPos,x
    
    ldy actorAnimation,x
    lda SprAnimAddrLo,y
    sta TEMP
    lda SprAnimAddrHi,y
    sta TEMP+1
    ldy actorAnimationPos,x
    iny
    lda (TEMP),y
    sta actorFrameDelay,x
    ; Preload next image if this is actor #0 or #1
    cpx #2
    bmi PreloadSpriteImage
    rts

PreloadSpriteImage:
    jsr GetPreloadPos
    tay
    lda (TEMP),y
    sta preloadSpriteImage,x
    lda preloadAddrMask,x
    eor ActorAddrMaskFlip,x
    sta preloadAddrMask,x
    rts
GetPreloadPos:
    lda actorAnimationPos,x
    clc
    adc #4
    ldy actorAnimation,x
    cmp SprAnimEnd,y
    bpl @loopedPos
    rts
@loopedPos:
    sec
    sbc SprAnimEnd,y
    clc
    adc #4
    rts
    
.MACRO ADVANCE_ANIMATION
    jsr AdvanceAnimationSub
.ENDMACRO

.MACRO RESTART_ANIMATION_IF_ENDED
.LOCAL dontRestartAnimation
    ldy actorAnimation,x
    lda actorAnimationPos,x
    cmp SprAnimEnd,y
    bne dontRestartAnimation
    lda #4
    sta actorAnimationPos,x
    ldy actorAnimation,x
    lda SprAnimAddrLo,y
    sta TEMP
    lda SprAnimAddrHi,y
    sta TEMP+1
    ldy #5
    lda (TEMP),y
    sta actorFrameDelay,x
dontRestartAnimation:
.ENDMACRO

AnimatedActorYieldSub:
    RESTART_ANIMATION_IF_ENDED
    jsr GetSpriteImageAtAnimationPos
    sta actorSpriteImage,x
    jmp YieldSub

;.include "blockactor.inc"

.CODE
GetMapPos:
    lda actorLoY,x
    and #$F0
    sta TEMP+0
    lda actorLoX,x
    lsr
    lsr
    lsr
    lsr
    ora TEMP+0
    tay
    rts

.SEGMENT "ACTORCODE"
GetMapPosRounded:
    lda actorLoY,x
    clc
    adc #$08
    and #$F0
    sta TEMP+0
    lda actorLoX,x
    clc
    adc #$08
    lsr
    lsr
    lsr
    lsr
    ora TEMP+0
    tay
    rts

SetMapPos:
    tya
    and #$0F
    asl
    asl
    asl
    asl
    sta actorLoX,x
    tya
    and #$F0
    sta actorLoY,x
    rts

;
; Checks if an actor of type Y exists at map positon Y
;
; Returns: C = 1 if actor found, C = 0 otherwise
;
CheckForActorTypeAtPosSub:
    tempX           = TEMP+0
    typeToLookFor   = TEMP+1
    xposToLookAt    = TEMP+2
    yposToLookAt    = TEMP+3
    stx tempX
    sta typeToLookFor
    tya
    pha
    mb xposToLookAt := A << 4
    tya
    mb yposToLookAt := A & #$F0
    ldx #MAX_ACTORS-1
    lda xposToLookAt
checkActorLoop:
    ldy actorAddrHi,x
    beq @UnusedActor
    cmp actorLoX,x
    beq matchedXpos
@UnusedActor:
    dex
    bpl checkActorLoop
    clc
    pla
    tay
    ldx tempX    
    rts
matchedXpos:
    lda yposToLookAt
    cmp actorLoY,x
    beq matchedYpos
    lda xposToLookAt
    dex
    bpl checkActorLoop
    clc
    pla
    tay
    ldx tempX
    rts
matchedYpos:
    lda typeToLookFor
    cmp actorType,x
    beq matchedAll
    lda xposToLookAt
    dex
    bpl checkActorLoop
    clc
    pla
    tay
    ldx tempX
    rts
matchedAll:
    sec
    pla
    tay
    ldx tempX
    rts

UpdateActors:
    ldx #MAX_ACTORS-1
@ActorLoop:
    lda actorAddrHi,x
    beq @SkipActor
    pha
    ADVANCE_ANIMATION
    pla
    jsr UpdateActor
    ;RESTART_ANIMATION_IF_ENDED ; Moved to YIELDing
@SkipActor:
    dex
    bpl @ActorLoop
    rts
    
UpdateActor:
    pha
    lda actorAddrLo,x
    pha
    lda actorFlags,x
    sta actorFlagsZP
    rts

.CODE
DrawActors:
    SWITCH_PRG #BANK_OAMCODE
    jsr SetSpriteImageToUpload

.repeat 4,i
    lda #(8*i)
    jsr DrawActorsPriorityLevel
.endrep
    rts

DrawActorsPriorityLevel:
    @priorityLevel = actorFlagsZP       ; unused by draw routine
    sta @priorityLevel
    ldx #MAX_ACTORS-1
@ActorLoop:
    lda actorAddrHi,x
    beq @NextActor
    lda actorFlags,x
    and #ACTOR_PRIORITY
    cmp @priorityLevel
    bne @NextActor
    stx TEMP+14
    jsr DrawActorAnimation
    ldx TEMP+14
@NextActor:
    dex
    bpl @ActorLoop
    rts

SetSpriteImageToUpload:
    ldx #1
@loop:
    lda actorSpriteImage,x
    tay
    lda ActorAddrMask+2,x
    eor preloadAddrMask,x
    cpy #NUM_IMAGES_BANK0-1
    bcs @dynamicImage
    lda #0
@dynamicImage:
    sta actorTileOffset,x

    dex
    bpl @loop
    rts

GetSpriteImageAtAnimationPos:
    ldy actorAnimation,x
    lda SprAnimAddrLo,y
    sta TEMP
    lda SprAnimAddrHi,y
    sta TEMP+1
    ldy actorAnimationPos,x
    lda (TEMP),y
    rts

GetActorDeltas:
	@animAddr           = TEMP+8
    @spriteImageIndex = TEMP+2
	@screenX            = TEMP+4
	@screenY            = TEMP+6
    @flags              = TEMP+12
    
    lda SprAnimAddrLo,y
    sta @animAddr
    lda SprAnimAddrHi,y
    sta @animAddr+1

    ; Get X
    lda actorFlags,x
    sta @flags
    asl
    bmi @Flipped_
    lda xAddTabLo,y
    ldy actorAnimationPos,x
    iny
    iny
    clc
    adc (@animAddr),y
    jmp :+
@Flipped_:
    lda xAddFlippedTabLo,y
    ldy actorAnimationPos,x
    iny
    iny
    sec
    sbc (@animAddr),y
:
    sta @screenX
    
    bmi @NegativeX
    lda #0
    beq :+
@NegativeX:
    lda #$FF
:
    sta @screenX+1
    
    ; Get Y
    iny
    lda (@animAddr),y
    sta @screenY
    bmi :+
    lda #0
    sta @screenY+1
    rts
:
    lda #$FF
    sta @screenY+1
    rts

DrawActorAnimation:
    @addr       = TEMP+0
    @spriteImageIndex = TEMP+2
    @animAddr           = TEMP+8
	@screenX            = TEMP+4
	@screenY            = TEMP+6
    @flags      = TEMP+12
    @tileOffset         = TEMP+15

    ; Use tile offset for actor #0 and #1 only
    lda #0
    cpx #2
    bpl :+
    lda actorTileOffset,x
:
    sta @tileOffset

    ldy actorAnimation,x
    bpl :+
    rts
:
    jsr GetActorDeltas
    
    lda actorLoX,x
    clc
    adc @screenX
    sta @screenX
    lda actorHiX,x
    and #1
    adc @screenX+1
    sta @screenX+1 
    
    lda actorLoY,x
    clc
    adc @screenY
    sta @screenY
    lda actorHiY,x
    and #1
    adc @screenY+1
    sta @screenY+1
    bne @NeedsClipping

    ; Clip if not (24 <= Y < 232)
    lda @screenY
    ;clc
    adc #(256 - 232)
    cmp #(24 + ( 256 - 232) )
    bcc @NeedsClipping    
    
    lda @screenX+1
    bne @NeedsClipping
    lda @screenX
    ; Clip if not (24 <= X < 232) 
    ;clc
    adc #(256 - 232)
    cmp #(24 + ( 256 - 232) )
    bcc @NeedsClipping
    
    ldy actorSpriteImage,x
    bne :+
    rts
:
	lda	SpriteImageAddrTabLo,y
	sta @addr
	lda	SpriteImageAddrTabHi,y
	sta @addr+1
    jmp DrawSpriteImageNoClipping

@NeedsClipping:
    ldy actorSpriteImage,x
	lda	SpriteImageAddrTabLo,y
	sta @addr
	lda	SpriteImageAddrTabHi,y
	sta @addr+1
    jmp DrawSpriteImage
    
FreeActor:
    lda #0
    sta actorAddrHi,x
    rts

FreeAllActors:
    ldx #MAX_ACTORS-1
    lda #0
:
    sta actorAddrHi,x
    dex
    bpl :-
    rts

.SEGMENT "ACTORCODE"    
FindFreeActor:
    ldx #0
:
    lda actorAddrHi,x
    beq @Found
    inx
    cpx #MAX_ACTORS
    bne :-
; If we ever get here, loop forever to indicate error
:
    jmp :-
    
@Found:
    rts


AddActor:
    pha
    jsr FindFreeActor
    pla
    jmp InitActor
        
ActorInitCodeTable:
    .word   0
    .word   GoodGuyInit-1
    .word   MonkeyInit-1
    .word   CrossHairInit-1
    .word   BananaInit-1
    .word   ThankyouInit-1

LoopForever:
    jmp LoopForever

InitActor:
    jsr InitActorSub
    lda actorAddrHi,x
    ;jsr UpdateActor
    rts

ActorBlockType:
.byte 0
.byte 0
.byte 0
.byte 0
.byte 0
.byte 0

ActorHeight:
.byte 0
.byte 0
.byte 0
.byte 0
.byte 0
.byte 0

InitActorSub:
    sta actorType,x
    lda #$FF
    sta actorAnimation,x
    tya
    sta actorMapPos,x
    and #$0F
    asl
    asl
    asl
    asl
    sta actorLoX,x
    tya
    and #$F0
    sta actorLoY,x

    lda #0
    sta actorFlags,x
    
    lda actorType,x
    asl
    tay
    lda ActorInitCodeTable,y
    sta actorAddrLo,x
    lda ActorInitCodeTable+1,y
    sta actorAddrHi,x

    ldy actorType,x
    lda ActorBlockType,y
    if zero
        rts
    fi
    
    pha
    
    txa
    tay
    ;jsr InitConstruction
    
    ldy actorType,x
    lda ActorHeight,y
    pha
    jsr GetMapPos
    pla
    cmp #1
    if equal
        pla
        ;jsr PutInMap1x1
        rts
    fi
    cmp #2
    if equal
        pla
        ;jsr PutInMap1x2
        rts
    fi
    pla
    rts

IncDirYSub:
    if ACTOR_FACING_RIGHT
        iny
        rts
    else
        dey
        rts
    fi

DecDirYSub:
    if ACTOR_FACING_RIGHT
        dey
        rts
    else
        iny
        rts
    fi

GoodGuyInit:
    .include "actors/goodguy.inc"
MonkeyInit:
    .include "actors/monkey.inc"
CrossHairInit:
    .include "actors/crosshair.inc"
BananaInit:
    .include "actors/banana.inc"
ThankyouInit:
    .include "actors/thankyou.inc"
