.SETCPU "6502X"

.include "common.h"

.ZEROPAGE
TEMP:                   .res 16
NMITEMP:                .res 7
delayMacroVar:          .res 1
frameCounter:           .res 2
RandSeed:               .res 2

.CODE

.align $100
Identity:
.repeat 256, i
.byte i
.endrep

.CODE
; Delays A:X clocks+overhead
; Time: 256*A+X+33 clocks (including JSR)
; Clobbers A. Preserves X,Y.
; Author: Bisqwit
.align $100
delay_256a_x_33_clocks:
        cmp #1                  ; +2; 2 cycles overhead
        bcs @do256              ; +2; 4 cycles overhead
        ; 0-255 cycles remain, overhead = 4
        txa                     ; +2; 6
        ;;;;;;;;;;;;;;;;
        ; 15 + JSR + RTS overhead for the code below. JSR=6, RTS=6. Total: 27. 6+27=33
        ;          ;    Cycles        Accumulator     Carry flag
        ;          ; 0  1  2  3  4       (hex)        0 1 2 3 4
        sec        ; 0  0  0  0  0   00 01 02 03 04   1 1 1 1 1
:       sbc #5     ; 2  2  2  2  2   FB FC FD FE FF   0 0 0 0 0
        bcs :-     ; 4  4  4  4  4   FB FC FD FE FF   0 0 0 0 0
        lsr a      ; 6  6  6  6  6   7D 7E 7E 7F 7F   1 0 1 0 1
        bcc :+     ; 8  8  8  8  8   7D 7E 7E 7F 7F   1 0 1 0 1
:       sbc #$7E   ;10 11 10 11 10   FF FF 00 00 01   0 0 1 1 1
        bcc :+     ;12 13 12 13 12   FF FF 00 00 01   0 0 1 1 1
        beq :+     ;      14 15 14         00 00 01       1 1 1
        bne :+     ;            16               01           1
:       rts        ;15 16 17 18 19   This loop from http://6502org.wikidot.com/software-delay
@do256: ; do 256 cycles.        ; 5 cycles done so far. C is set from CMP
        sbc #1                  ; 2 cycles
        pha                     ; 3 cycles
         lda #(34*2-1)          ; 2 cycles
         ;                      ;12 cycles done so far
:        sec                    ; 2 cycles  (sec is only needed
         sbc #2                 ; 2 cycles   to make loop 7 cycles)
         bcs :-                 ; 3 cycles for taken branch
         ;                      ; -1 cycles for untaken branch
         ;12 + 34*7 - 1 = 249 done so far, 7 missing
        pla                        ; 4 cycles
        bcc delay_256a_x_33_clocks ; 3 cycles ; C is unset from SBC

MemCopySub:
    @src        = TEMP
    @dst        = TEMP+2
    @loopCnt    = TEMP+4
    sta @loopCnt
    ldy #0
@Loop:
    lda (@src),y
    sta (@dst),y
    iny
    dex
    bne @Loop
    inc @src+1
    inc @dst+1
    dec @loopCnt
    bmi :+
    bne @Loop
:
    rts

ClearRAM:
    ldx #0
    txa
:
    sta $000,x
    ;sta $100,x
    sta $200,x
    sta $300,x
    sta $400,x
    sta $500,x
    sta $600,x
    sta $700,x
    inx
    bne :-
    ; Init seed for TSW RNG
    lda #$47
    sta RandSeed
    rts

;
; Tausworthe RNG
; s = (((s<<1)>>8) & 0xFF) ^ (s & 0xFF) | (s>>15) ^ (s>>8))<<8
; s = (s & 0xFF) | ((s<<1) & 0xFF) ^ (s>>8)
Rand:
        lda RandSeed
        asl
        lda RandSeed+1
        rol
        eor RandSeed
        sta RandSeed
        lda RandSeed+1
        asl
        lda #0
        rol
        eor RandSeed+1
        sta RandSeed+1
        lda RandSeed
        asl
        eor RandSeed+1
        sta RandSeed+1
        rts
