.SETCPU "6502X"

.include "common.h"
.include "actor.h"
.include "gfx.h"

.IMPORTZP   frameCounter

spritePage			= $200
drawList            = $100

.ZEROPAGE
drawListPosR:           .res 1
drawListPosW:           .res 1
currentSpritePtr:       .res 1
attributeRowUpdated:    .res 8
paletteShadow:          .res 25
attributeShadow:        .res 64

.SEGMENT "PPUCODE"
TileData:
.REPEAT 256,i
.byte i
.ENDREP
.INCBIN "levels/jungle/jungle.chr" 

SpriteTileData:
.INCBIN "dragon.chr"

.SEGMENT "OAMCODE"
.include "dragon.spr"

.SEGMENT "PPUCODE"
;
; Format of draw list
;
; 0: Data length
; 1: 4 if inc-by-32, 0 if inc-by-1
; 2: PPUADDR_HI
; 3: PPUADDR_LO
; 4: ...N data bytes...
;
ProcessDrawList:
    @tempX  = NMITEMP+2
    @addr   = NMITEMP+0
    lda #>@UnrolledCopyLoop
    sta @addr+1
    tsx
    stx @tempX
    ldx #0 ;drawListPosR
    dex
    txs

    jmp @DoOneTransfer

.align $100
@UnrolledCopyLoop:
.repeat 64,i
pla
sta $2007
.endrep
@DoOneTransfer:
    pla
    tay
    beq @EndOfList
    ; branchaddr = 256-4*num_bytes = NOT(4*num_bytes)+1+256 = NOT(4*num_bytes)+1
    lda @NumBytesToAddress,y
    sta @addr
    pla
    sta $2000
    pla
    sta $2006
    pla
    sta $2006
    jmp (@addr)
    
;:
;    pla
;    sta $2007
;    dey
;    bne :-
;    jmp @DoOneTransfer

@EndOfList:
    tsx
    inx
    stx drawListPosR
    ldx @tempX
    txs
    rts

@NumBytesToAddress:
.repeat 65,i
.byte <(256-4*i)
.endrep    

.CODE
ClearDrawList:
    lda #0
    sta drawListPosW
    sta drawListPosR
    ldx #8
:
.repeat 16,i
    sta drawList+i*8,x
.endrep
    dex
    bpl :-
    rts

.SEGMENT "PPUCODE"
WriteAttributeShadow:
    ldx #7
@Loop:
    lda attributeRowUpdated,x
    bne @RowUpdated
    dex
    bpl @Loop
    rts
 @RowUpdated:
    lda #$23
    sta $2006
    lda @attributeShadowOffs,x
    sta $2006
    and #$38
    tay
    lda attributeShadow+0,y
    sta $2007
    lda attributeShadow+1,y
    sta $2007
    lda attributeShadow+2,y
    sta $2007
    lda attributeShadow+3,y
    sta $2007
    lda attributeShadow+4,y
    sta $2007
    lda attributeShadow+5,y
    sta $2007
    lda attributeShadow+6,y
    sta $2007
    lda attributeShadow+7,y
    sta $2007
    dex
    bpl @Loop
    rts
@attributeShadowOffs:
.byte $C0,$C8,$D0,$D8,$E0,$E8,$F0,$F8

.CODE
ClearAttributeRowUpdated:
    lda #0
    sta attributeRowUpdated+0
    sta attributeRowUpdated+1
    sta attributeRowUpdated+2
    sta attributeRowUpdated+3
    sta attributeRowUpdated+4
    sta attributeRowUpdated+5
    sta attributeRowUpdated+6
    sta attributeRowUpdated+7
    rts
    
ClearVRAM:
    lda #$00
    sta $2006
    sta $2006
    ldy #64
    ldx #0
@PageLoop:
:
    sta $2007
    dex
    bne :-
    dey
    bne @PageLoop
    rts

.SEGMENT "PPUCODE"
;
; Y = starting tile
; X = number of tiles
; C = CHR bank
; @ChrPtr = tile data
;
UploadTiles:
    @ChrPtr = TEMP
    @temp = TEMP+2
    sty @temp
    lda #$00
    rol
    asl @temp
    rol
    asl @temp
    rol
    asl @temp
    rol
    asl @temp
    rol
    sta $2006
    lda @temp
    sta $2006
@TilesLoop:
    ldy #0
:
    lda (@ChrPtr),y
    sta $2007
    iny
    cpy #16
    bne :-
    lda @ChrPtr
    clc
    adc #16
    sta @ChrPtr
    lda @ChrPtr+1
    adc #0
    sta @ChrPtr+1
    dex
    bne @TilesLoop
    rts

UploadBgCHR:
    @ChrPtr = TEMP
    lda #>TileData
    sta @ChrPtr+1
    lda #<TileData
    sta @ChrPtr
    ldy #0
    ldx #132
    clc
    jmp UploadTiles
    
UploadSpriteCHR:
    @ChrPtr = TEMP
    lda #>SpriteTileData
    sta @ChrPtr+1
    lda #<SpriteTileData
    sta @ChrPtr
    ldx #0
    ldy #0
    sec
    jmp UploadTiles

.CODE
UploadCHR:
    SWITCH_PRG #BANK_PPUCODE
    jsr UploadBgCHR
    jsr UploadSpriteCHR
    rts
    
SetPalette:
    ldy #0
:
    lda @palette,y
    sta paletteShadow,y
    iny
    cpy #25
    bne :-    
    rts

@palette:
.byte $11
.byte $07,$1A,$0A
.byte $08,$28,$30
.byte $2D,$10,$30
.byte $08,$17,$28

.byte $28,$17,$08
.byte $1D,$11,$30
.byte $28,$08,$1D
.byte $08,$18,$30

HideUnusedSprites:
    ldy currentSpritePtr
    beq @skipLoop
    lda #$F0
:
    sta spritePage,y
    iny
    iny
    iny
    iny
    bne :-
@skipLoop:
    ldy #(4*4)
    sty currentSpritePtr
    rts

.SEGMENT "OAMCODE"

;
; A = flags 
;
DrawSpriteImageNoClipping:
	@addr       = TEMP+0
	@screenX    = TEMP+4
	@screenY    = TEMP+6
	@flags      = TEMP+12
    @tileOffset = TEMP+15
	
    ;sta @flags
    
    dec currentSpritePtr
	lda	@addr
	sec
	sbc	currentSpritePtr
	sta	@addr
    bcs :+
    dec @addr+1
:
	
    ldy	currentSpritePtr
    lda (@addr),y
    iny
    ; If currentSpritePtr has wrapped around, do no more!
    bne :+
    sty currentSpritePtr
    rts
:
    ;;;
    tax
    ; Adjust X so we don't overflow currentSpritePtr: max X = 64 - (Y/4) = -(Y/4) + 64
    stx TEMP+13
    tya
    lsr
    lsr
    eor #$FF
    adc #1+64
    cmp TEMP+13
    bcs @noSpriteLimiting
    tax
@noSpriteLimiting:
    ;;;

    lda @flags
    and #$40
    bne @Flipped
	clc
@Loop:
	lda	(@addr),y
	adc	@screenY
	sta	spritePage,y
	iny
	lda	(@addr),y
    adc @tileOffset
	sta	spritePage,y
	iny
	lda	(@addr),y
    eor @flags
	sta	spritePage,y
	iny
	lda	(@addr),y
	adc	@screenX
	sta	spritePage,y
	iny
	dex
	bne	@Loop
    sty currentSpritePtr
	rts

@Flipped:
@LoopFlipH:
	lda	(@addr),y
    clc
	adc	@screenY
	sta	spritePage,y
	iny
	lda	(@addr),y
    adc @tileOffset
	sta	spritePage,y
	iny
	lda	(@addr),y
    eor @flags
	sta	spritePage,y
	iny
    lda @screenX
    sec
    sbc (@addr),y
    sta spritePage,y
	iny
	dex
	bne	@LoopFlipH
    sty currentSpritePtr
	rts

;
; A = flags 
;
DrawSpriteImage:
	@addr       = TEMP+0
	@screenX    = TEMP+4
	@screenY    = TEMP+6
	@flags      = TEMP+12
    @tileOffset = TEMP+15
	
    ;sta @flags
    
	lda	@addr
	sec
	sbc	currentSpritePtr
	sta	@addr
    bcs :+
	;lda	@addr+1
	;sbc	#0
	;sta	@addr+1
    dec @addr+1
:
	
    ldy	currentSpritePtr
    ; If currentSpritePtr has wrapped around, do no more!
    bne :+
    rts
:
    ;;;
    lda (@addr),y
    tax
    ; Adjust X so we don't overflow currentSpritePtr: max X = 64 - (Y/4) = -(Y/4) + 64
    stx TEMP+13
    tya
    lsr
    lsr
    eor #$FF
    adc #1+64
    cmp TEMP+13
    bcs @noSpriteLimiting
    tax
@noSpriteLimiting:
    ;;;
    
    lda @addr
    clc
    adc #1
    sta @addr
    lda @addr+1
    adc #0
    sta @addr+1
    
    lda @flags
    and #$40
    bne @Flipped
	
@Loop:
	lda	(@addr),y
	clc
	adc	@screenY
	sta	spritePage,y
	lda	#0
	adc	@screenY+1
	;bmi	@OutsideY
	bne	@OutsideY
	iny
	lda	(@addr),y
    clc
    adc @tileOffset
	sta	spritePage,y
	iny
	lda	(@addr),y
    eor @flags
	sta	spritePage,y
	iny
	lda	(@addr),y
	clc
	adc	@screenX
	sta	spritePage,y
	lda	#0
	adc	@screenX+1
	;bmi	@OutsideX
	bne	@OutsideX
	iny
	dex
	bne	@Loop
    sty currentSpritePtr
	rts

@OutsideY:
	lda	#$F0
	sta	spritePage,y
	iny
	iny
	iny
	iny
	dex
	bne	@Loop
    sty currentSpritePtr
	rts
	
@OutsideX:
	lda	#$F0
	sta	spritePage-3,y
	iny
	dex
	bne	@Loop
    sty currentSpritePtr
	rts

    
@Flipped:
@LoopFlipH:
	lda	(@addr),y
	clc
	adc	@screenY
	sta	spritePage,y
	lda	#0
	adc	@screenY+1
	bmi	@OutsideYFlipH
	bne	@OutsideYFlipH
	iny
	lda	(@addr),y
    clc
    adc @tileOffset
	sta	spritePage,y
	iny
	lda	(@addr),y
    eor @flags
	sta	spritePage,y
	iny

    lda @screenX
    sec
    sbc (@addr),y
    sta spritePage,y
    lda @screenX+1
    sbc #0
    ;bmi @OutsideXFlipH
    bne @OutsideXFlipH
    
	iny
	dex
	bne	@LoopFlipH
    sty currentSpritePtr
	rts

@OutsideYFlipH:
	lda	#$F0
	sta	spritePage,y
	iny
	iny
	iny
	iny
	dex
	bne	@LoopFlipH
    sty currentSpritePtr
	rts
	
@OutsideXFlipH:
	lda	#$F0
	sta	spritePage-3,y
	iny
	dex
	bne	@LoopFlipH
    sty currentSpritePtr
	rts

.macro WRITE_CHRPAGE page
    ldy #11
@loop:
    lda (NMITEMP),y
    tax
.repeat 16,i
    lda page+$100*i,x
    sta $2007
.endrep
    dey
    bpl @loop
.endmacro

.SEGMENT "DRAGONCHR"
.align $100
dragonCHR1:
;.incbin ".rhc";

dragonCHR2:
;.incbin ".rhc"

dragonCHR3:
;.incbin ".rhc"

WriteChrPageDragon1:
    WRITE_CHRPAGE dragonCHR1
    rts

WriteChrPageDragon2:
    WRITE_CHRPAGE dragonCHR2
    rts

WriteChrPageDragon3:
    WRITE_CHRPAGE dragonCHR3
    rts

WriteGraphicsBank:
    lda frameCounter
    and #1
    tax
    ldy preloadSpriteImage,x
    lda TileRefAddrLo,y
    sta NMITEMP
    lda TileRefAddrHi,y
    sta NMITEMP+1
  
@WriteSpriteChr:
    lda ActorAddrMask,x
    eor preloadAddrMask,x
    sta NMITEMP+2
    lda #$01
    asl NMITEMP+2
    rol
    asl NMITEMP+2
    rol
    asl NMITEMP+2
    rol
    asl NMITEMP+2
    rol
    sta $2006
    lda NMITEMP+2
    sta $2006
    lda TilesPage,y
    cmp #1
    bne :+
    jmp WriteChrPageDragon1
:
    cmp #2
    bne :+
    jmp WriteChrPageDragon2
:
    jmp WriteChrPageDragon3

.include "dragon_tilesref.inc"    

.CODE
ActorAddrMask:
.byte $D0
.byte $E8
.byte $DC
.byte $F4

ActorAddrMaskFlip:
.byte $0C
.byte $1C
