#! /usr/bin/env python

import sys, os
import array
from PIL import Image
import imp
import string
from collections import OrderedDict

filler = 1
num4xTilesToReserve = 4

class LevelData(object):
    def __init__(self):
        self.blocks = OrderedDict()
        self.blocks[(0,4,8,12, 0x00)] = len(self.blocks) # Ice
        self.blocks[(0,4,8,12, 0x55)] = len(self.blocks) # Stone
        self.tiles4x = []
        self.tiles1x = []
        self.bgDict = OrderedDict()

    def addTiles(self, nameTable, attributeTable, tiles, doSolids):
        for y in range(15):
            for x in range(16):
                aPos = 8*(y>>1)+(x>>1)
                aShift = 2*(x & 1) + 4*(y & 1)
                attribute = (attributeTable[aPos]>>aShift) & 3
                solid = attribute==2    # Temporary hack! Change to reading .map file
                if solid == doSolids:
                    for tile in [nameTable[32*2*y+2*x],nameTable[32*2*y+2*x+1],nameTable[32*(2*y+1)+2*x],nameTable[32*(2*y+1)+2*x+1]]:
                        if tile not in tiles:
                            tiles.append(tile)

    def addBG(self, filename):
        byteData = array.array('B', [])
        byteData.fromfile(open(filename, 'rb'), os.path.getsize(filename))
        nameTable = byteData[0:(32*30)]
        attributeTable = byteData[(32*30):1024]
        # First do one pre-pass only processing solids, where we reserve 4x the space to allow quick bitmask ops for ice connections
        tiles = []        
        self.addTiles(nameTable, attributeTable, self.tiles4x, True)
        # Now do second pass, adding in the rest (1x) tiles
        self.addTiles(nameTable, attributeTable, self.tiles1x, False)
        self.bgDict[filename.replace('.nam','')] = (nameTable,attributeTable)

    def createRemapping(self):
        # All tiles we reserved 4x space for need to reside in the lowest part of the tiletable, so remap them!
        self.remapNewToOld = []
        self.remapOldToNew = [0xFF]*256
        for i,tileIndex in enumerate(sorted(self.tiles4x)):
            self.remapOldToNew[tileIndex] = len(self.remapNewToOld)+4*num4xTilesToReserve #.append(len(self.remapNewToOld))
            self.remapNewToOld.append(tileIndex)
            self.remapNewToOld.append(-1)
            self.remapNewToOld.append(-1)
            self.remapNewToOld.append(-1)
        # Now the 1x space tiles...
        for i,tileIndex in enumerate(sorted(self.tiles1x)):
            self.remapOldToNew[tileIndex] = len(self.remapNewToOld)+4*num4xTilesToReserve #.append(len(self.remapNewToOld))
            self.remapNewToOld.append(tileIndex)

    def writeRemappedBG(self, filename, f):
        nameTable,attributeTable = self.bgDict[filename]
        f.write('bgMap_%s:\n' % os.path.split(filename)[1])
        for y in range(15):
            f.write('.byte ')
            for x in range(16):
                aPos = 8*(y>>1)+(x>>1)
                aShift = 2*(x & 1) + 4*(y & 1)
                attribute = (attributeTable[aPos]>>aShift) & 3
                tileUL = self.remapOldToNew[nameTable[32*2*y+2*x]]
                tileUR = self.remapOldToNew[nameTable[32*2*y+2*x+1]]
                tileLL = self.remapOldToNew[nameTable[32*(2*y+1)+2*x]]
                tileLR = self.remapOldToNew[nameTable[32*(2*y+1)+2*x+1]]
                block = (tileUL,
                         tileUR,
                         tileLL,
                         tileLR,
                         attribute|(attribute<<2)|(attribute<<4)|(attribute<<6) )
                if block not in self.blocks:
                    self.blocks[block] = len(self.blocks)
                f.write('$%02X' % self.blocks[block])
                if x == 15:
                    f.write('\n')
                else:
                    f.write(',')

    def writeTablesCHR(self, f):
        labels = ['bgTileUL','bgTileUR','bgTileLL','bgTileLR','bgAttribute']
        for i,label in enumerate(labels):
            f.write('%s:\n.byte ' % label)
            for block,index in self.blocks.iteritems():
                f.write('$%02X' % block[i])
                if index == len(self.blocks)-1:
                    f.write('\n')
                else:
                    f.write(',')

    def writeRemappedCHR(self, filename):
        chrDataOld = array.array('B', [])
        chrDataOld.fromfile(open(filename + ".chr", 'rb'), os.path.getsize(filename + ".chr"))
        chrDataNew = array.array('B', [])
        for i in range(len(self.remapNewToOld)):
            if self.remapNewToOld[i] != -1:
                for j in range(16):
                    oldIndex = self.remapNewToOld[i]
                    #print "oldIndex = ",oldIndex
                    chrDataNew.append(chrDataOld[oldIndex*16+j])
        chrDataNew.tofile(open(filename + "_remapped.chr", 'wb'))

#
# Run as stand-alone program
#        
if __name__ == "__main__":
    if len(sys.argv) < 3:
        print "Usage: convertbg.py <commonchr> <nametable#1> ... <nametable#N>"
        sys.exit(2)
    f = sys.stdout
    
    chrname = sys.argv[1]

    levelData = LevelData()
    for filename in sys.argv[2:]:
        levelData.addBG(filename)

    # Remove tiles in tiles1x if they are also present in tile4x
    for tile in levelData.tiles4x:
        if tile in levelData.tiles1x:
            levelData.tiles1x.remove(tile)

    levelData.createRemapping()
   
    # Create new remapped chr data... but do not repeat 4x tiles here!
    levelData.writeRemappedCHR(chrname)
   
    # Finally, create blockmap
    for filename,t in levelData.bgDict.iteritems():
        levelData.writeRemappedBG(filename, f)

    # Create common CHR tables
    levelData.writeTablesCHR(f)

