#! /usr/bin/env python

import sys, os, copy
import array
from PIL import Image
import imp
import string
from collections import OrderedDict

sys.path.append('/home/michel/tilificator/trunk')
from sprite import *
from tile import TileTable,TileSettings
from import_raw_tiles import RawFormats,writeTiles

paletteSize = 16

#
# Run as stand-alone program
#        
if __name__ == "__main__":
    if len(sys.argv) < 2:
        print "Usage: nesifyimage.py <image file>"
        sys.exit(2)
    f = sys.stdout

    filename = sys.argv[1]

    si = SpriteImage(filename)
    imageWidth,imageHeight = si.width,si.height

    # Make copy of color data and then remove high bits to avoid duplicate tiles
    colorData = copy.copy(si.data)
    for i in range(len(si.data)):
        si.data[i] &= 3

    basename = filename.split('.')[0]
    tileMap = []
    colorMap = []
    settings = TileSettings()
    settings.colorSize = 4
    tileWidth = settings.tileWidth
    tileHeight = settings.tileHeight
    tileTable = TileTable(settings)
    # Add blank tile at #0
    tileTable.addTile(array.array('B', [0 for n in range(tileWidth*tileHeight)]))
    for y in range(imageHeight/tileHeight):
        for x in range(imageWidth/tileWidth):
            # Read tile at (x*tileWidth,y*tileHeight)
            tileData = tileTable.cutTile(si, x*tileWidth, y*tileHeight, flipH = False, flipV = False)
            paletteData = colorData[si.width*y*tileHeight+x*tileWidth]/paletteSize
            # Get tileID from tileTable
            tileID,flipH,flipV = tileTable.findTile(tileData)
            if tileID < 0:
                tileID = tileTable.numTiles
                tileTable.addTile(tileData)
            tileMap.append(tileID)
            colorMap.append(paletteData)     
    writeTiles(basename + '.chr', tileTable.data, tileTable.numTiles, RawFormats['NES'])

    # Split tileMap into 32x30 nametables
    numScreensHorizontal = imageWidth/256
    numScreensVertical = imageHeight/240
    for j in range(numScreensVertical):
        for i in range(numScreensHorizontal):
            nameTable = array.array('B', [0 for n in range(32*32)])
            for y in range(30):
                for x in range(32):
                    stride = 32*numScreensHorizontal
                    offset = stride*(32*j)+(32*i)
                    nameTable[32*y+x] = tileMap[offset+stride*y+x]
                    # Attribute
                    attribute = colorMap[offset+stride*y+x]
                    xx = x>>1
                    yy = y>>1
                    aPos = 8*(yy>>1)+(xx>>1)
                    aShift = 2*(xx & 1) + 4*(yy & 1)
                    nameTable[32*30+aPos] |= (attribute<<aShift)
            if sum(nameTable):
                nameTable.tofile(open('%s_%d_%d.nam'%(basename,i,j),'wb'))


