.SETCPU "6502X"

.EXPORT NMI,IRQ,MainCode
.EXPORTZP updateState

.include "actor.h"
.include "common.h"
.include "input.h"
.include "gfx.h"
.include "map.h"
.include "maploader.h"
.include "soundcode.h"
.include "serial.h"

.macro SET_SPRITEPAGE_VALID
    sec
    ror spritePageValid
.endmacro

.macro SET_SPRITEPAGE_INVALID
    lsr spritePageValid
.endmacro

.macro SET_DRAWLIST_VALID
    sec
    ror drawListValid
.endmacro

.macro SET_DRAWLIST_INVALID
    lsr drawListValid
.endmacro

.ZEROPAGE
pauseState:             .res 1
updateState:            .res 1
spritePageValid:        .res 1
drawListValid:          .res 1
spriteHitOccurred:      .res 1
NMI_Done:               .res 1
frameCounterOld:        .res 2
;framesElapsed:          .res 1
;RunLua:                 .res 1
LuaFlags:               .res 1
waterSurfaceY:          .res 1
currentMap:             .res 1

.CODE

InitSpriteZero:
    ; Sprite hit
    lda #165
    sta spritePage      ; Y coord
    lda #$CF
    sta spritePage+1    ; tile number
    lda #$00
    sta spritePage+2    ; Flags
    lda #241
    sta spritePage+3    ; X coord
    lda #4
    sta currentSpritePtr
    rts
    
MainCode:
    lda #$80
    sta $4017
    sei ;cli
    jsr ClearRAM
    jsr ClearVRAM
    
    jsr SoundInit
    
    lda #3
    sta currentMap
    jsr LoadNextMap

    lda #1
    sta pauseState
    lda #$20
    sta $2006
    lda #$00
    sta $2006
  
    lda #$1E
    sta $2001
    lda #$80
    sta $2000
 
	lda	#0
	sta	frameCounter
	sta	frameCounter+1
    
    ;jsr FullMapUpdate
 
    lda #$80
    sta $2000

    SWITCH_PRG #BANK_ACTORCODE
    ; Add monkey
    lda #ACTOR_TYPE_MONKEY
    ldy #$49
    jsr AddActor

    jsr init_serial_57600

    MSG_BYTE "R",#0
    MSG_BYTE "R",#1
    MSG_BYTE "R",#2

@loop:
:
    lda NMI_Done
    bpl :-
    jsr ReadZapper ;ReadJoypads
    ;jsr init_serial_57600
    ;lda #$1F
    ;sta $2001
    ;MSG_BYTE "X",#42
    ;MSG_BYTE "Y",#47
    ;jsr SendMessage
    ;lda #$1E
    ;sta $2001
    jsr UpdateAnimation

    jsr SoundUpdate

    lda #0
    sta NMI_Done
    jmp @loop


BigDelay:
    ldx #0
:
    dex
    bne :-
    rts

SendMessage:
    lda #'G'
    jsr write_serial_57600
    lda #'l' 
    jsr write_serial_57600
    lda #'o'
    jsr write_serial_57600
    lda #'b'
    jsr write_serial_57600
    lda #'a'
    jsr write_serial_57600
    lda #'l'
    jsr write_serial_57600
    lda #'G'
    jsr write_serial_57600
    lda #'a'
    jsr write_serial_57600
    lda #'m'
    jsr write_serial_57600
    lda #'e'
    jsr write_serial_57600
    lda #'J'
    jsr write_serial_57600
    lda #'a'
    jsr write_serial_57600
    lda #'m'
    jsr write_serial_57600
    lda #'2'
    jsr write_serial_57600
    lda #'0'
    jsr write_serial_57600
    lda #'1'
    jsr write_serial_57600
    lda #'4'
    jsr write_serial_57600
    lda #13
    jsr write_serial_57600
    lda #10
    jsr write_serial_57600
    rts

UpdateAnimation:
    ;lda #$9F
    ;sta $2001
    jsr ExchangeJoypads
    SWITCH_PRG #BANK_ACTORCODE
    jsr UpdateActors
    jsr ExchangeJoypads
	lda	frameCounter
	clc
	adc	#1
	sta	frameCounter
	lda	frameCounter+1
	adc	#0
	sta	frameCounter+1
    SET_SPRITEPAGE_INVALID
    ;lda #$5F
    ;sta $2001
    ;jsr DrawSpriteBlocks
    jsr DrawActors
    ;lda #$3F
    ;sta $2001
    jsr HideUnusedSprites
    SET_SPRITEPAGE_VALID
    ;lda #$1E
    ;sta $2001
    rts

UpdateDrawMap:
    jsr ClearAttributeRowUpdated
    SET_DRAWLIST_INVALID
    jsr ClearDrawList
    jsr DrawMap
    SET_DRAWLIST_VALID
    ;jsr ResetEnvUpdatedFlag
    ;jsr FindSpriteBlocks
    rts

DoNothing:
    rts
   
FullMapUpdate:
    SET_DRAWLIST_INVALID
    lda #$00
    sta $2000
    jsr WriteEntireMapToPPU
	lda #$80
    sta $2000
    jsr ClearDrawList
    SET_DRAWLIST_VALID
    rts

LoadNextMap:
    SET_DRAWLIST_INVALID
    sei
    lda #$00
    sta $2000
    sta $2001
    lda #0
    sta currentMap
    jsr LoadMap
    inc currentMap
    lda #$00
    sta $2000
    sta $2001
    jsr UploadCHR
    lda #$00
    sta $2006
    sta $2006
    ;jsr UploadEditCursorCHR
    jsr SetPalette
    jsr InitSpriteZero
    jsr ClearDrawList

    jsr FullMapUpdate
    cli
    lda #$88
    sta $2000
    rts
    
IRQ:
    rti

NMI:
    pha
    txa
    pha
    tya
    pha

    lda $2002

    lda #$08
    sta $2000
    lda #0
    sta $2001

    NMI_SWITCH_PRG #BANK_PPUCODE
    
    WRITE_PALETTE_SHADOW
    jsr WriteAttributeShadow
    bit drawListValid
    bpl :+    
    jsr ProcessDrawList
:

    lda #$F0
    sta spritePage      ; sprite zero hit Y coord

    ; Hide unused sprites
    lda #$F0
    sta spritePage+4
    sta spritePage+8
    sta spritePage+12

    bit spritePageValid
    bpl :+
    lda #0
    sta $2003
    lda #>spritePage
    sta $4014
:

    ; Enable screen to get normal dot crawl pattern
    lda #$FE
    sta $2001

    lda $2002
    lda #$00
    sta $2006
    sta $2006
    
    NMI_SWITCH_PRG #BANK_DRAGONCHR
    
    ; Wait for end of vblank if-and-only-if sprite hit occurred in previous frame
;    bit $2002
;    bvc @NoWait
;:
;    bit $2002
;    bvs :-
@NoWait:

    ; DEBUG: delay 11 scanlines to make visible 
;    ldy #250
;:
;    dey
;    bne :-
    DELAY 59+41 ; 2560+41 ; +40.666666...
    ; Best so far: +41 (always affects fire on stone)
    ; Delay 1.75 scanline before disabling screen to avoid messing up dot crawl pattern
    ; (0.75 scanline is to avoid sprite OAM refresh flicker bug)
    ldy #43 ; #40
:
    dey
    bne :-
    lda $2002
    lda #$08
    sta $2000
    lda #$00
    sta $2001
    jsr WriteGraphicsBank

    
   
;    ; Turn screen on in second half of scanline to avoid OAM Refresh flicker bug
;    ldy #26
;:
;    dey
;    bne :-
;    nop

    ; start scrolling at line 17
;    nop
;    nop
;    NOP
    SCROLLX = 0
    SCROLLY = 10
    ;lda $2002
    lda #<(SCROLLY>>3)
    sta $2006
    ldy #<(SCROLLY & $FF)
    ldx #<(SCROLLX & $FF)
    lda #<((SCROLLX>>3)|((SCROLLY>>3)<<5))
    sty $2005
    stx $2005
    sta $2006

    ldx #$18
    stx $2001
    
    DELAY 18 ;16 ; 17 ;DELAY 80
    
    bit drawListValid
    bpl :+
    jsr ClearDrawList
:
	
    lda #$88
    sta $2000
    sta NMI_Done
   
    NMI_SWITCH_PRG currentBank
    pla
    tay
    pla
    tax
    pla
    rti
