.SETCPU "6502X"

.include "common.h"

.include "map.h"

BLOCK_BEEN_VISITED      = %10000000
BLOCK_STATIONARY        = %01000000
BLOCK_CONNECTED_LEFT    = %00000001
BLOCK_CONNECTED_RIGHT   = %00000010
BLOCK_CONNECTED_UP      = %00000100
BLOCK_CONNECTED_DOWN    = %00001000
BLOCK_CONNECTED_ALL     = BLOCK_CONNECTED_RIGHT+BLOCK_CONNECTED_LEFT+BLOCK_CONNECTED_UP+BLOCK_CONNECTED_DOWN

ENVIRONMENT_UPDATED     = %10000000
ENVIRONMENT_WATER       = %00010000

MAX_SPRITEBLOCKS        = 8

.include "gfx.h"
.include "level.h"

.ZEROPAGE
bgMapPtr:               .res 2

.CODE

WriteEntireMapToPPU:
    lda #0
    sta $2001
    jsr ResetEnvUpdatedFlag
    ldx #15
@RowLoop:
    txa
    pha

    asl
    asl
    asl
    asl
    tax
    ldy #16
:
    lda envMap,x
    ora #ENVIRONMENT_UPDATED
    sta envMap,x
    inx
    dey
    bne :-
    jsr ClearAttributeRowUpdated
    jsr ClearDrawList
    jsr DrawMap
    SWITCH_PRG #BANK_PPUCODE
    jsr WriteAttributeShadow
    jsr ProcessDrawList
    jsr ResetEnvUpdatedFlag
    pla
    tax
    dex
    bpl @RowLoop
    rts

ResetEnvUpdatedFlag:
    ldx #0
:
    lda envMap,x
    and #<~ENVIRONMENT_UPDATED
    sta envMap,x
    inx
    bne :-
    rts

;
; Searches the map for tiles needing updates, and writes those to the draw list
; 
DrawMap:
    @tempX                  = TEMP+1
    @sectionStartX          = TEMP+2
    @sectionEndX            = TEMP+3
    @orMaskUL               = TEMP+4
    @orMaskUR               = TEMP+5
    @orMaskLL               = TEMP+6
    @orMaskLR               = TEMP+7
    @bitMaskLowerOrUpper    = TEMP+9
    @bitMaskPPUAddrLo       = TEMP+10
    @blockTypeAttribute     = TEMP+11
    @andMaskAttribute       = TEMP+12
    @blockMapType           = TEMP+13

    ldx #0
@DrawMapLoop:
    lda envMap,x
    bmi @NeedsUpdate
    inx
    cpx #$F0
    bne @DrawMapLoop
    rts

@NeedsUpdate:
    stx @sectionStartX

    ; First obtain length of section needing update
:
    lda envMap,x
    bpl @EndOfSection
    inx
    bne :-
@EndOfSection:
    stx @sectionEndX
    txa
    tay
    
    ; Draw upper half of tiles
    lda #0
    sta @bitMaskLowerOrUpper
    sta @bitMaskPPUAddrLo
    ldy @sectionStartX
    jsr DrawTilesSection
    
    ; Draw lower half of tiles
    lda #$80
    sta @bitMaskLowerOrUpper
    lda #$20
    sta @bitMaskPPUAddrLo
    ldy @sectionStartX
    jsr DrawTilesSection

    ldx @sectionEndX
    cpx #$F0
    bne @DrawMapLoop
    rts

    
;
; In: X = map position START, Y = map position END
;
DrawTilesSection:
    @tempY                  = TEMP+1
    @sectionStartX          = TEMP+2
    @sectionEndX            = TEMP+3
    @orMaskUL               = TEMP+4
    @orMaskUR               = TEMP+5
    @orMaskLL               = TEMP+6
    @orMaskLR               = TEMP+7
    @bitMaskLowerOrUpper    = TEMP+9
    @bitMaskPPUAddrLo       = TEMP+10
    @blockTypeAttribute     = TEMP+11
    @andMaskAttribute       = TEMP+12
    @blockMapType           = TEMP+13

    ldx drawListPosW
    lda #0              ; inc-by-1
    sta drawList+1,x
    lda mapPosToNameTableAddrUpperHi,y
    sta drawList+2,x
    lda mapPosToNameTableAddrUpperLo,y
    ora @bitMaskPPUAddrLo
    sta drawList+3,x
    inx
    inx
    inx
    inx
@SectionLoop:
    lda @bitMaskLowerOrUpper
    sta TEMP+14
    sty @tempY
    stx TEMP+0
    
    jsr @adjustFG

    jmp @DrawBG

    bit @bitMaskLowerOrUpper
    bvs @DrawBG
    lda blockMapType,y
    beq @DrawBG
    bmi @DrawBG

    lda blockMapFlags,y
    and #BLOCK_CONNECTED_LEFT+BLOCK_CONNECTED_RIGHT+BLOCK_CONNECTED_UP+BLOCK_CONNECTED_DOWN    
    tax
    lda orMasksUL,x
    sta @orMaskUL
    lda orMasksUR,x
    sta @orMaskUR
    lda orMasksLL,x
    sta @orMaskLL
    lda orMasksLR,x
    sta @orMaskLR
    lda blockMapType,y
    ;cmp #2 ;BLOCK_TYPE_STATIONARY
    ;bne :+
    lda (bgMapPtr),y
    jmp @Draw
;:
;    sec
;    sbc #1 ;BLOCK_TYPE_ICE
;    jmp @Draw
   
@DrawBG:
    ; BG should show no connections
    lda #0
    sta @orMaskUL
    sta @orMaskUR
    sta @orMaskLL
    sta @orMaskLR
    lda (bgMapPtr),y
@Draw:
    tay
    ldx TEMP+0
    bit @bitMaskLowerOrUpper
    bmi @bgLower
    lda @orMaskUL
    ora bgTileUL,y
    sta drawList,x
    inx
    lda @orMaskUR
    ora bgTileUR,y
    sta drawList,x
    inx
    jmp :+
@bgLower:
    lda @orMaskLL
    ora bgTileLL,y
    sta drawList,x
    inx
    lda @orMaskLR
    ora bgTileLR,y
    sta drawList,x
    inx
:
    
    lda bgAttribute,y
    sta @blockTypeAttribute
    ldy @tempY
    lda andMaskAttribute,y
    sta @andMaskAttribute
    ; Update attribute shadow
    ; offs = 8*(row>>1) + (column>>1)
    ; attributeShadow[offs] = (attributeShadow[offs] & ~((0x3<<(column & 0x2))<<(row & 0x2))) | ((attribute<<(column & 0x2))<<(row & 0x2))
    lda mapPosToAttributeTableAddr,y
    tay
    lda @andMaskAttribute
    and @blockTypeAttribute
    sta @blockTypeAttribute
    lda @andMaskAttribute
    eor #$FF
    and attributeShadow,y
    ora @blockTypeAttribute
    sta attributeShadow,y
    tya
    lsr
    lsr
    lsr
    tay
    lda #1
    sta attributeRowUpdated,y
    
    ldy @tempY
    iny
    cpy @sectionEndX
    beq :+;bne @SectionLoop
    lda TEMP+14
    sta @bitMaskLowerOrUpper
    jmp @SectionLoop
:

    lda #0
    sta drawList,x
    lda drawListPosW
    stx drawListPosW
    tax
    lda @sectionEndX
    sec
    sbc @sectionStartX
    asl
    sta drawList,x
    rts

@adjustFG:
    ; drawFGbelow  if: (block[x][y+1] is non-empty) and (block[x][y+1] is shifted up) and lowerSection
    ;   => load block[x][y+1], bgLower = false
    lda blockMapType+16,y
    beq :+
    bmi :+
    lda #0 ;fallingDelta
    beq :+
    ldx blockMapIndex+16,y
    bmi :+
    lda #1 ;constructionFlags,x
    ;and #CONSTRUCTION_STABLE
    bne :+
    bit @bitMaskLowerOrUpper
    bpl :+
    tya
    clc
    adc #16
    tay
    lda #$00
    sta @bitMaskLowerOrUpper
    rts
:

    ; drawFG       if: (block[x][y] is non-empty)
    ;   => load block[x][y],   bgLower = (block[x][y] is shifted up) or lowerSection
    bit @bitMaskLowerOrUpper
    bmi :+
    lda blockMapType,y
    beq :+
    bmi :+
    lda #0 ;fallingDelta
    beq :+
    ldx blockMapIndex,y
    lda #1 ;constructionFlags,x
    ;and #CONSTRUCTION_STABLE
    bne :+
    lda #$80
    sta @bitMaskLowerOrUpper
    rts
:

    ; if (block[x][y] is shifted up) => load BG
    lda blockMapType,y
    bmi :+
    lda #0 ;fallingDelta
    beq :+
    ldx blockMapIndex,y
    lda #1 ;constructionFlags,x
    ;and #CONSTRUCTION_STABLE
    bne :+
    lda @bitMaskLowerOrUpper
    ora #$40
    sta @bitMaskLowerOrUpper
:
    rts

.align $100
mapPosToNameTableAddrUpperHi:
.repeat 256, i
.byte >($2000 + 32*2*(i>>4) + 2*(i & $0F))
.endrep
mapPosToNameTableAddrUpperLo:
.repeat 256, i
.byte <($2000 + 32*2*(i>>4) + 2*(i & $0F))
.endrep

;.align $100
;mapPosToNameTableAddrLowerHi:
;.repeat 256, i
;.byte >($2000 + 32*(2*(i>>4)+1) + 2*(i & $0F))
;.endrep
;mapPosToNameTableAddrLowerLo:
;.repeat 256, i
;.byte <($2000 + 32*(2*(i>>4)+1) + 2*(i & $0F))
;.endrep

.align $100
mapPosToAttributeTableAddr:
.repeat 256, i
.byte 8*((i>>4)>>1) + ((i & $0F)>>1)
.endrep
andMaskAttribute:
.repeat 256, i
.byte ($3<<(((i & $0F) & $1)*2))<<(((i>>4) & $1)*4)
.endrep

orMasksUL:
.byte 0,1,0,1,2,3,2,3,0,1,0,1,2,3,2,3
orMasksUR:
.byte 0,0,1,1,2,2,3,3,0,0,1,1,2,2,3,3
orMasksLL:
.byte 0,1,0,1,0,1,0,1,2,3,2,3,2,3,2,3
orMasksLR:
.byte 0,0,1,1,0,0,1,1,2,2,3,3,2,2,3,3
