.include "serial.h"

; Serial I/O at 57600 bits/sec
;
; Assumes connection through inverting level converter,
; like MAX232 or FTDI cable, with D0 going to Tx and Strobe
; going to Rx.
;
; Uses stack and register A only, and doesn't mind page
; crossing (uses subroutines instead of loops).
;
; Doesn't check for framing errors on receive (missing
; stop bit).

; Set to 1 if using PAL NES
.ifndef SERIAL_PAL
	SERIAL_PAL = 1
.endif

; Which controller to receive data on
.ifndef SERIAL_PORT
	SERIAL_PORT = $4017 ; second controller
.endif

; Initializes serial
init_serial_57600:
	sec
	lda #$FF
	jmp :+

; Writes byte to serial
; In: A = byte to send
; Preserved: X, Y
write_serial_57600:
	clc
:       jsr @bit        ; start
	jsr @bit        ; bit 0
	jsr @bit        ; bit 1
	jsr @bit        ; bit 2
	jsr @bit        ; bit 3
	jsr @bit        ; bit 4
	jsr @bit        ; bit 5
	jsr @bit        ; bit 6
	jsr @bit        ; bit 7
	sec             ; 2 stop bit
	nop             ; 4
	nop 
@bit:                   ; 6 jsr
	pha             ; 3
	rol a           ; 2
	and #$01        ; 2
	sta $4016       ; 4
	pla             ; 4
	ror a           ; 2
.if SERIAL_PAL = 0
	nop             ; 2
.endif
	rts             ; 6

; Waits for and reads byte from serial. Max 22 cycles between
; JSRs to serial_read, or data corruption will occur.
; Out: A = byte received
; Preserved: X, Y
read_serial_57600:      ; 6 JSR
	lda #1          ; 2
:       bit SERIAL_PORT ; 4
	beq :-          ; 3
			; average 3.5 cycles after start bit
	pha             ; 21+4 delay for start bit
	pla
	pha
	pla
	pha
	pla
	jsr @first      ; bit 0
	jsr @bit        ; bit 1
	jsr @bit        ; bit 2
	jsr @bit        ; bit 3
	jsr @bit        ; bit 4
	jsr @bit        ; bit 5
	jsr @bit        ; bit 6
	nop             ; bit 7
@first: nop
	nop
@bit:                   ; 6 JSR
	pha             ; 3
.if SERIAL_PAL = 0
	nop             ; 2
.endif
	lda #1          ; 2
	eor SERIAL_PORT ; 4
	lsr a           ; 2
	pla             ; 4
	ror a           ; 2
	rts             ; 6


